from .field import CRUDField
from .action import CRUDAction

# For a new basic set of CRUD operations you could just do

# from whyqd.metamodel.crud.base import CRUDBase
# from whyqd.metamodel.models import ItemModel

# item = CRUDBase[ItemModel]()
